module Main where
import qualified Network.Socket as Soc
import qualified Directory as D
import qualified Data.Map as M
import qualified Data.Set as S
import qualified Data.List as L
import qualified Control.Concurrent.Chan as C
import qualified Control.Concurrent.MVar as V
import qualified Control.Concurrent as T
import qualified Data.ByteString.Lazy.UTF8 as U
import qualified Data.ByteString.Lazy as BS
import qualified Data.ByteString as SBS
import qualified Random as R
import qualified System.Process as P
import qualified System.Posix.Signals as PS
import qualified System.Posix.Process as PP
import qualified System.Posix.Types as PT
import Text.Regex.Posix
import System.IO
import Control.Monad
import Control.Arrow

data PlayState = Stop | Pause | Play deriving (Eq)
data State = State {
	playlist :: [U.ByteString],
	history :: [U.ByteString],
	modes :: S.Set String,
	pstate :: PlayState,
	stop :: Task,
	plpause :: Task,
	sid :: Int
	}
type Task = State -> IO State
type Worker = Task -> IO ()

getWorker :: IO Worker
getWorker = do
		chan <- C.newChan
		let result = C.writeChan chan
		tasks <- C.getChanContents chan
		T.forkIO $ foldM_ (\state task -> task state) State {
			playlist = [],
			history = [],
			modes = S.empty,
			pstate = Stop,
			stop = return,
			plpause = next result,
			sid = 0
			} tasks
		return result

create_sock :: IO Soc.Socket
create_sock = do
		catch ( D.removeFile address ) $ const $ return ()
		socket <- Soc.socket Soc.AF_UNIX Soc.Stream 0
		Soc.bindSocket socket $ Soc.SockAddrUnix address
		Soc.listen socket 10
		return socket
	where
		address = ".hclueplay"

operate :: Soc.Socket -> IO ()
operate socket = getWorker >>= (forever . acceptConn socket)

handleConn :: Worker -> Handle -> IO ()
handleConn worker = hGetContents >=> handleInput worker . map (chomp . U.fromString) . lines

chomp :: U.ByteString -> U.ByteString
chomp = fst . U.break (== '\n')

terminated :: Worker -> Int -> Task
terminated worker osid newstate = if sid newstate == osid + 1 then next worker (newstate { pstate = Stop }) else return newstate

start :: Worker -> String -> Task
start worker filename state = do
	putStrLn $ '>':filename
	hFlush stdout
	if filename =~ "\\.(ogg|waw|mp3)" then startMy worker filename state else startMPL worker filename state

startMy :: Worker -> String -> Task
startMy worker filename state = do
	pid <- PP.forkProcess $ PP.executeFile "myplay" True [filename] Nothing
	T.forkIO (PP.getProcessStatus True False pid >> (worker $ terminated worker $ sid state))
	return state {
		pstate = Play,
		stop = \state -> catch (PS.signalProcess PS.killProcess pid) (const $ return ()) >> return state {
			pstate = Stop,
			stop = return,
			plpause = next worker,
			sid = sid state + 1
			},
		plpause = let
				action sig st ac = \state -> catch (PS.signalProcess sig pid) (const $ return ()) >> return state {
					pstate = st,
					plpause = ac
					}
				pause = action PS.softwareStop Pause play
				play = action PS.continueProcess Play pause
			in pause,
		sid = sid state + 1
		}

startMPL :: Worker -> String -> Task
startMPL worker filename state = do
	(inp, out, err, pid) <- P.runInteractiveProcess "/home/vorner/bin/qmplayer" ((if "novideo" `S.member` modes state then ["-novideo"] else []) ++ [filename]) Nothing Nothing
	hSetBuffering inp NoBuffering
	reader out
	reader err
	waiter pid worker
	return state {
		pstate = Play,
		stop = \state -> (catch (hPutChar inp 'q') $ const $ return ()) >> return state {
			pstate = Stop,
			stop = return,
			plpause = next worker,
			sid = sid state + 1
			},
		plpause = \state -> (catch (hPutChar inp ' ') $ const $ return ()) >> return state {
			pstate = if pstate state == Play then Pause else Play
			},
		sid = sid state + 1
		}
	where
		reader handle = T.forkIO (hGetContents handle >>= putStr)
		waiter pid worker = T.forkIO $ P.waitForProcess pid >> (worker $ terminated worker $ sid state)

shuffle :: State -> IO State
shuffle state = replicateM (length $ playlist state) (R.randomIO :: IO Int) >>= \rnd -> return $ state {
	playlist = map snd $ L.sort $ zip rnd $ playlist state
	}

maybeShuffle :: State -> IO State
maybeShuffle state = if "random" `S.member` modes state && sid state `mod` 40 == 0 then shuffle state else return state

preload :: State -> IO State
preload state = T.forkIO (mapM_ (readFile . U.toString >=> writeFile "/dev/null") $ take 20 $ playlist state) >> return state

maybePreload :: State -> IO State
maybePreload state = if "preload" `S.member` modes state && sid state `mod` 20 == 1 then putStrLn "Preload" >> preload state else return state

next :: Worker -> Task
next worker state = case playlist state of
	[] -> return state
	song:pl -> stop state state >>= start worker (U.toString song) >>= maybeShuffle >>= maybePreload >>= \ns -> return ns {
		history = take 100 (song:history ns),
		playlist = if "repeat" `S.member` modes ns then pl ++ [song] else pl
		}

prev :: Worker -> Task
prev worker state = case history state of
	[] -> return state
	song:hist -> stop state state >>= start worker (U.toString song) >>= \ns -> return ns {
		history = hist
		}

handleInput :: Worker -> [U.ByteString] -> IO ()
handleInput worker (line:rest) = putStrLn (U.toString line) >> command >>= handleInput worker
	where
		(cmd, arg) = break (==' ') $ U.toString line
		command = M.findWithDefault (putStrLn ("Unknow command " ++ cmd) >> return rest) cmd commands
		action :: Task -> IO [U.ByteString]
		action f = worker f >> return rest
		mod f = action $ return . f
		mode f = mod (\state -> state { modes = f (dropWhile (==' ') arg) $ modes state })
		args :: ([U.ByteString] -> State -> IO State) -> IO [U.ByteString]
		args f = let (vals, cmds) = break (\str -> (str == U.fromString "") || (U.take 1 str /= U.fromString ">")) rest in (worker $ f (map (U.drop 1) vals)) >> return cmds
		internal :: (State -> Task) -> IO [U.ByteString]
		internal f = action (\state -> f state state)
		clear state = state { playlist = [] }
		append files state = preload $ state { playlist = playlist state ++ files }
		commands :: M.Map String (IO [U.ByteString])
		commands = M.fromList $ [
			("clear", mod clear),
			("append", args append),
			("load", args (\files -> append files . clear)),
			("setmode", mode S.insert),
			("unsetmode", mode S.delete),
			("shuffle", action shuffle),
			("next", action $ next worker),
			("playpause", internal plpause),
			("stop", internal stop),
			("prev", action $ prev worker)
			]
handleInput _ _ = return ()

acceptConn :: Soc.Socket -> Worker -> IO ()
acceptConn masterS worker = Soc.accept masterS >>= (`Soc.socketToHandle` ReadMode) . fst >>= T.forkIO . handleConn worker >> return ()

main :: IO ()
main = Soc.withSocketsDo $ create_sock >>= operate
